from urllib.parse import urljoin, quote_plus
from leapcell.version import VERSION

def multi_urljoin(*parts):
    return urljoin(parts[0], "/".join(quote_plus(part.strip("/"), safe="/") for part in parts[1:]))

def build_header(token: str):
    headers = {
        "Authorization": f"Bearer {token}",
        "User-Agent": "Leapcell Python-Client/{}".format(
            VERSION,
        ),
    }
    return headers


from typing import List, Tuple, Dict, Union, Any
import base64
from leapcell.version import VERSION
import urllib.parse
from functools import reduce
from leapcell.utils import multi_urljoin, build_header
import json

FILE_UPLOAD_TIMEOUT = 600
FILE_UPLOAD_MAX_SIZE = 1024 * 1024 * 5



class LeapcellFile(object):
    def __init__(self, obj: Dict[str, Any]) -> None:
        id = obj.get("id", "")
        self._id = id
        link = obj.get("link", "")
        self._link = link
        meta = obj.get("meta", {})
        self._meta = meta

    def id(self) -> str:
        return self._id

    def link(self) -> str:
        return self._link

    def width(self) -> int:
        return self._meta.get("width", 0)

    def height(self) -> int:
        return self._meta.get("height", 0)

    def __getitem__(self, __name: str) -> Any:
        if __name == "id":
            return self.id()
        elif __name == "link":
            return self.link()
        elif __name == "width":
            return self.width()
        elif __name == "height":
            return self.height()
        else:
            raise KeyError("invalid key {}".format(__name))
        
    def __str__(self) -> str:
        return "LeapcellFile(id={}, link={}, width={}, height={})".format(
            self.id(), self.link(), self.width(), self.height()
        )

    def __repr__(self) -> str:
        return json.dumps(self.tojson(), ensure_ascii=False)

    def tojson(self):
        return {
            "id": self._id,
            "link":self._link,
            "meta": {
                "width": self._meta.get("width", 0),
                "height": self._meta.get("height", 0),
            }
        }


from typing import Dict, Union, Any, Optional
from leapcell.table_meta import TableMeta
from leapcell.http_client import HTTPClient
from leapcell.const import TableFieldType
import json


class Record:
    """Leapcell Record Instance

    Raises:
        KeyError: _description_

    Returns:
        _type_: _description_
    """

    _data: Dict[str, Any]
    _record_id: Optional[str]
    _table_meta: TableMeta
    _field_type: TableFieldType

    def __init__(
        self,
        requester: HTTPClient,
        record_id: Optional[str] = None,
        fields: Optional[Dict[str, Any]] = None,
        create_time: Optional[int] = None,
        update_time: Optional[int] = None,
    ) -> None:
        self._data = {}
        if fields is not None:
            for field, item in fields.items():
                self._data[field] = item
            self._record_id = record_id
        self._new_data: Dict[str, Any] = {}
        self._requester = requester
        self._create_time = create_time
        self._update_time = update_time
        return

    def __getitem__(self, key: str) -> Any:
        return self._data[key]

    def __setitem__(self, key: str, value: Any) -> None:
        data = value
        self._data[key] = data
        self._new_data[key] = data

    def __delitem__(self, key: str) -> None:
        del self._data[key]
        del self._new_data[key]

    def __iter__(self):
        return iter(self._data)

    def __len__(self):
        return len(self._data)

    def data(self) -> Dict[str, Any]:
        return self._data

    def updated(self) -> Dict[str, Any]:
        return self._new_data

    @property
    def record_id(self) -> str | None:
        return self._record_id

    @property
    def id(self) -> Union[None, str]:
        return self._record_id

    @property
    def create_time(self) -> int | None:
        return self._create_time

    @property
    def update_time(self) -> int | None:
        return self._update_time

    def __repr__(self) -> str:
        return json.dumps(self.tojson(), ensure_ascii=False)

    def __str__(self) -> str:
        return "<record_id: {}, data: {}, create_time: {}, update_time: {}>".format(
            self._record_id, self._data, self._create_time, self._update_time
        )

    def save(self) -> None:
        update_values = dict()
        for field_name, value in self.updated().items():
            update_values[field_name] = value

        if not self._record_id:
            return

        self._requester.update_record(
            record_id=self._record_id,
            data={
                "fields": update_values,
            },
        )
        return

    def delete(self) -> None:
        self._requester.delete_record(
            record_id=self._record_id,
        )
        return

    def toJSON(self):
        return {
            "record_id": self._record_id,
            "data": self._data,
            "create_time": self._create_time,
            "update_time": self._update_time,
        }

    def tojson(self):
        return self.toJSON()

    def get(self, key: str, default: Any = None) -> Any:
        return self._data.get(key, default)



from leapcell.const import TableFieldType
from leapcell.table_meta import TableMeta
from typing import Dict, Union, Any, List, Tuple, Optional
from leapcell.field_meta import FieldMeta
from leapcell.http_client import HTTPClient
from leapcell.field_item import FieldMgr, BaseItem
from leapcell.record import Record
from functools import reduce
from leapcell.file import LeapcellFile
import json

support_op = [
    "eq",
    "gt",
    "gte",
    "lt",
    "lte",
    "neq",
    "contain",
    "in",
    "not_in",
    "is_null",
    "not_null",
]


def is_list_of_strings(lst):
    return (
        bool(lst)
        and not isinstance(lst, str)
        and all(isinstance(elem, str) for elem in lst)
    )


class LeapcellFilter(object):
    def __init__(self, **args):
        type_ = args.get("type", "eq")
        self.filter = {"filter": {"type": type_}}

        if type_ in support_op:
            self.filter["filter"].update(
                {"field": args["field"], "value": args["value"]}
            )
        elif type_ == "and":
            self.filter["filter"].update({"fields": args["fields"]})
        elif type_ == "or":
            self.filter["filter"].update({"fields": args["fields"]})
        elif type_ == "not":
            self.filter["filter"].update({"fields": args["fields"]})
        else:
            raise KeyError("Filter type: {0} does not exist".format(type_))

    def __and__(self, x):
        if self.filter["filter"]["type"] == "and":
            self.filter["filter"]["fields"].append(x)
            return self
        return LeapcellFilter(type="and", fields=[self, x])

    def __or__(self, x):
        if self.filter["filter"]["type"] == "or":
            self.filter["filter"]["fields"].append(x)
            return self
        return LeapcellFilter(type="or", fields=[self, x])

    def __invert__(self, x):
        if self.filter["filter"]["type"] == "not":
            self.filter["filter"]["fields"].append(x)
            return self
        return LeapcellFilter(type="not", fields=[self, x])

    def __str__(self) -> str:
        return "<filter: {}>".format(self.filter)

    @staticmethod
    def build_filter(filter_obj):
        filter = filter_obj.filter["filter"]
        if filter["type"] in ["and", "or", "not"]:
            filter = filter.copy()  # make a copy so we don't overwrite `fields`
            filter["fields"] = [
                LeapcellFilter.build_filter(f) for f in filter["fields"]
            ]
        return filter


# leapcell tool class
class LeapcellOrder(object):
    def __init__(self, field: str, order: str = "desc") -> None:
        self._field = field
        self._order = order

    def __str__(self) -> str:
        return f"{self._field} {self._order}"

    def __repr__(self) -> str:
        return f"{self._field} {self._order}"

    def get_order(self) -> Tuple[str, str]:
        return (self._field, self._order)


class KaithQuery(object):
    """Leapcell Query class

    Args:
        object (_type_): _description_
    """

    def __init__(
        self,
        requester: HTTPClient,
        fields: List[str] = [],
        filter: Optional[LeapcellFilter] = None,
        orders: List[Tuple[str, str]] = [],
        offset: int = 0,
        limit: int = 20,
        aggr: Optional[str] = None,
    ) -> None:
        self.fields = fields
        self._filter = filter
        self._orders = orders
        self._offset = offset
        self._limit = limit
        self._aggr = aggr
        self._requester = requester

    def __str__(self) -> str:
        return "<filter: {}, orders: {}, offset: {}, limit: {}, aggr: {}>".format(
            self._filter,
            self._orders,
            self._offset,
            self._limit,
            self._aggr,
        )

    def where(self, filter: Optional[LeapcellFilter] = None):
        """where condition

        Args:
            filter (Optional[LeapcellFilter], optional): Filter condition. Defaults to None.

        Raises:
            TypeError: filter must be a LeapcellFilter object, you can use bracket to build a filter

        Returns:
            _type_: _description_
        """
        if isinstance(filter, bool):
            raise TypeError(
                "filter must be a LeapcellFilter object, you can use bracket to build a filter"
            )
        if not filter:
            raise TypeError(
                "filter must be a LeapcellFilter object, you can use bracket to build a filter"
            )
        elif self._filter is None:
            self._filter = filter
        else:
            self._filter = filter & self._filter
        return self

    def order_by(
        self,
        orders: List[Tuple[str, str] | LeapcellOrder]
        | LeapcellOrder
        | Tuple[str, str] = [],
    ):
        """order by

        Args:
            orders (List[Tuple[str, str]  |  LeapcellOrder] | LeapcellOrder | Tuple[str, str], optional): _description_. Defaults to [].

        Raises:
           ValueError: order_by must be a list of tuple (name, 'desc') or table[name].desc() or table[name].asc()

        Returns:
            _type_: _description_
        """
        if isinstance(orders, LeapcellOrder):
            self._orders.append(orders.get_order())
            return self
        if isinstance(orders, list) and len(orders) == 0:
            return self
        if (
            isinstance(orders, tuple)
            and len(orders) == 2
            and isinstance(orders[0], str)
            and isinstance(orders[1], str)
        ):
            self._orders.append(orders)
            return self

        for order in orders:
            if isinstance(order, tuple):
                self._orders.append(order)
            elif isinstance(order, LeapcellOrder):
                self._orders.append(order.get_order())
            else:
                raise ValueError(
                    "order_by must be a list of tuple (name, 'desc') or table[name].desc() or table[name].asc()"
                )
        return self

    def limit(self, limit: int = 50):
        """limit

        Args:
            limit (int, optional): _description_. Defaults to 50.

        Returns:
            _type_: _description_
        """
        self._limit = limit
        return self

    def take(self, limit: int = 50):
        return self.limit(limit)

    def skip(self, offset: int = 0):
        return self.offset(offset)

    def offset(self, offset: int = 0):
        """offset

        Args:
            offset (int, optional): _description_. Defaults to 0.

        Returns:
            _type_: _description_
        """
        self._offset = offset
        return self

    def select(self, fields: List[str] = []):
        """field to select

        Args:
            fields (List[str], optional): _description_. Defaults to [].

        Returns:
            _type_: _description_
        """
        self.fields = fields
        return self

    def search(
        self,
        query: str,
        search_fields: List[str] = [],
        boost_fields: Dict[str, int] = {},
    ) -> List[Record]:
        """search records by keyword

        Args:
            query (str): search keyword
            fields (List[str], optional): search fields. Defaults to [].
            boost_fields (Dict[str, int], optional): boost fields. Defaults to {}.
        """

        resp = self._search(query, search_fields, boost_fields)
        if not resp or not resp["records"]:
            return []
        return [
            Record(
                requester=self._requester,
                record_id=record["record_id"],
                fields=record["fields"],
                create_time=record.get("create_time", None),
                update_time=record.get("update_time", None),
            )
            for record in resp["records"]
        ]

    def query(self):
        """execute query

        Returns:
            _type_: Record instance list
        """
        resp = self._query()
        if resp["records"] is None:
            return []
        return [
            Record(
                requester=self._requester,
                record_id=record["record_id"],
                fields=record["fields"],
                create_time=record.get("create_time", None),
                update_time=record.get("update_time", None),
            )
            for record in resp["records"]
        ]

    def first(self):
        """get the first record

        Returns:
            _type_: _description_
        """
        self._limit = 1
        resp = self._query()
        if resp["records"] is None:
            return None
        if len(resp["records"]) == 0:
            return None

        return Record(
            requester=self._requester,
            record_id=resp["records"][0]["record_id"],
            fields=resp["records"][0]["fields"],
            create_time=resp["records"][0].get("create_time", None),
            update_time=resp["records"][0].get("update_time", None),
        )

    def update(self, values: Dict[str, Any]):
        """execute update

        Args:
            values (Dict[str, Any]): _description_

        Returns:
            _type_: _description_
        """
        return self._update(values)

    def delete(self):
        """execute delete

        Returns:
            _type_: _description_
        """
        return self._delete()

    def count(self):
        """execute count

        Returns:
            _type_: _description_
        """
        return self._count()

    def _get_filter(self, filter: Union[LeapcellFilter, None] = None) -> Dict | None:
        query_filter: Optional[Dict] = None

        def gen_filter(f: Dict[str, Any]):
            op = None
            val = f["value"]
            if f["type"] == "gt":
                op = "gt"
            elif f["type"] == "gte":
                op = "gte"
            elif f["type"] == "lt":
                op = "lt"
            elif f["type"] == "lte":
                op = "lte"
            elif f["type"] == "neq":
                op = "neq"
            elif f["type"] == "eq":
                op = "eq"
            elif f["type"] == "in":
                op = "in"
            elif f["type"] == "not_in":
                op = "not_in"
            elif f["type"] == "not_null":
                op = "not_null"
            elif f["type"] == "is_null":
                op = "is_null"
            elif f["type"] == "contain":
                op = "contain"
            else:
                raise KeyError("Filter type: {0} does not exist".format(f["type"]))
            field = f["field"]
            return {
                "val": val,
                "op": op,
                "field": field,
            }

        if filter is not None and isinstance(filter, LeapcellFilter):
            fs = LeapcellFilter.build_filter(filter)
            if fs["type"] in ["and", "or", "not"]:
                filters = []
                for f in fs["fields"]:
                    fitem = gen_filter(f)

                    filters.append(fitem)
                query_filter = {
                    "filterType": fs["type"],
                    "filters": filters,
                }
            else:
                fitem = gen_filter(fs)
                query_filter = fitem
        elif isinstance(filter, dict):
            ff = self._condition2filter(filter)
            return self._get_filter(ff)

        return query_filter

    def _gen_order(self, orders: List[Tuple[str, str]]) -> List[Dict]:
        sortByCol = []
        if orders is not None:
            for order in orders:
                if not isinstance(order, tuple):
                    raise TypeError("order must be tuple, like [('view_id',)]")
                if len(order) > 2:
                    raise ValueError(
                        "invalid order, is should be like [('xx', 'desc')]"
                    )
                if len(order) == 2 and order[1] not in ["desc", "asc"]:
                    raise ValueError("invalid order type, should in ['desc', 'asc']")
                order_type = "DESC"
                if len(order) == 2 and order[1] == "asc":
                    order_type = "ASC"
                field_id = order[0]
                sortByCol.append({"field": field_id, "sortType": order_type})
        return sortByCol

    def _query(
        self,
    ) -> Dict | None:
        fields = self.fields
        filter = self._filter
        orders = self._orders
        offset = self._offset
        limit = self._limit
        aggr = self._aggr

        query_filter: Optional[Dict] = self._get_filter(filter)
        sortByCol = self._gen_order(orders)

        req: Dict[str, Any] = {
            "limit": limit,
            "offset": offset,
            "filter": query_filter,
        }
        if aggr:
            req["aggr"] = aggr
        if fields:
            req["fields"] = fields
        if sortByCol:
            req["orders"] = sortByCol

        resp = self._requester.get_records(req)

        return resp

    def _update(
        self,
        values: Dict[str, Any],
    ) -> Optional[int]:
        filter = self._filter
        record_values = dict()
        for key, value in values.items():
            record_values[key] = value
        query_filter: Optional[Dict] = self._get_filter(filter)

        resp = self._requester.update_records(
            {
                "filter": query_filter,
                "fields": record_values,  # data transform, like datetime to int
            }
        )

        if not resp:
            return None

        return resp["affect_count"]

    def _search(
        self,
        query: str,
        search_fields: List[str] = [],
        boost_fields: Dict[str, int] = {},
    ) -> Optional[Dict[str, Any]]:
        fields = self.fields
        filter = self._filter
        orders = self._orders
        offset = self._offset
        limit = self._limit

        query_filter: Optional[Dict] = self._get_filter(filter)
        sortByCol = self._gen_order(orders)

        req: Dict[str, Any] = {
            "query": query,
            "search_fields": search_fields,
        }
        if query_filter:
            req["filter"] = query_filter
        if sortByCol:
            req["orders"] = sortByCol
        if fields:
            req["fields"] = fields
        if offset:
            req["offset"] = offset
        if limit:
            req["limit"] = limit
        if boost_fields:
            req["boost_fields"] = boost_fields

        resp = self._requester.search(req)
        if not resp:
            return None

        return resp

    def _delete(
        self,
    ) -> Optional[int]:
        filter = self._filter

        query_filter: Optional[Dict] = self._get_filter(filter)

        resp = self._requester.delete_records(
            {
                "filter": query_filter,
            }
        )

        if not resp:
            return None

        return resp["affect_count"]

    def _count(
        self,
        distinct: Optional[bool] = None,
    ) -> Optional[int]:
        filter = self._filter

        query_filter: Optional[Dict] = self._get_filter(filter)

        params: Dict[str, Any] = {
            "filter": query_filter,
            "metric": {
                "field": "*",
                "aggr": "count",
            },
        }
        if distinct:
            params["metric"]["condition"] = "distinct"
        resp = self._requester.aggr_record(params)

        if not resp or not resp["metric"]:
            return None

        return resp["metric"]["value"]

    def _condition2filter(
        self, conditions: Optional[Dict[str, Any] | LeapcellFilter]
    ) -> Optional[LeapcellFilter]:
        filter: Optional[LeapcellFilter] = None
        if conditions is not None and isinstance(conditions, Dict):
            filters = [LeapcellField(field) == val for field, val in conditions.items()]
            filter = reduce(lambda x, y: x and y, filters)
        elif conditions is not None and isinstance(conditions, LeapcellFilter):
            filter = conditions
        return filter


class LeapcellField(object):
    """Leapcell Field class

    Args:
        object (_type_): _description_
    """

    def __init__(self, field: str):
        self.field = field

    def __eq__(self, other):
        return LeapcellFilter(field=self.field, value=other, type="eq")

    def eq_(self, other):
        return self.__eq__(other)

    def __gt__(self, other):
        return LeapcellFilter(field=self.field, value=other, type="gt")

    def gt_(self, other):
        return self.__gt__(other)

    def __ge__(self, other):
        return LeapcellFilter(field=self.field, value=other, type="gte")

    def ge_(self, other):
        return self.__ge__(other)

    def __lt__(self, other):
        return LeapcellFilter(field=self.field, value=other, type="lt")

    def lt_(self, other):
        return self.__lt__(other)

    def __le__(self, other):
        return LeapcellFilter(field=self.field, value=other, type="lte")

    def le_(self, other):
        return self.__le__(other)

    def __ne__(self, other):
        return LeapcellFilter(field=self.field, value=other, type="neq")

    def neq_(self, other):
        return self.__ne__(other)

    def contain(self, other):
        return LeapcellFilter(field=self.field, value=other, type="contain")

    def __contains__(self, other):
        return self.contain(other)

    def in_(self, value: List[Any]):
        return LeapcellFilter(field=self.field, value=value, type="in")

    def not_in(self, value: List[Any]):
        return LeapcellFilter(field=self.field, value=value, type="not_in")

    def is_null(self):
        return LeapcellFilter(field=self.field, value=None, type="is_null")

    def not_null(self):
        return LeapcellFilter(field=self.field, value=None, type="not_null")

    # TODO: to complete these operations
    def order_by(self, order_type: str = "desc"):
        return LeapcellOrder(field=self.field, order=order_type)

    def desc(self):
        return LeapcellOrder(field=self.field, order="desc")

    def asc(self):
        return LeapcellOrder(field=self.field, order="asc")


class LeapcellTable(object):
    """Leapcell table instance

    Args:
        object (_type_): _description_
        resource (str): leapcell resource name, you can find it in leapcell with format leapcell.io/{username}/{repo_name}, resource all have the same format as {username}/{repo_name}
        token (str): leapcell token
        table_id (str): ltable id, you can find it in leapcell with format leapcell.io/{username}/{repo_name}/table/{table_id}, table_id all have the same format as {username}/{repo_name}/table/{table_id}
        base_url (str): leapcell base url, default is https://api.leapcell.io
        field_type (str): field_type is the type of the field, it can be "id" or "name". Defaults to "name". if it's "id", the field name will be "13145252145", "13145252147" ...; if it's "name", the field name will be "field-0", "field-1", "field-2", ...,
        version (str): leapcell api version, default is v1

    Raises:
        KeyError: if field not found in table, raise KeyError

    Returns:
        LeapcellTable: LeapcellTable instance
    """

    _field_name_type: TableFieldType

    def __init__(
        self,
        repository: str,
        api_key: str,
        table_id: str,
        base_url: str,
        name_type: str = "name",
        version: str = "v1",
    ) -> None:
        if name_type == "name":
            self._field_name_type = TableFieldType.NAME
        else:
            self._field_name_type = TableFieldType.ID
        self._resource = repository
        self._requster = HTTPClient(
            api_key=api_key,
            base_url=base_url,
            resource=repository,
            table_id=table_id,
            version=version,
            name_type=name_type,
        )
        self._table_id = table_id

    def __repr__(self) -> str:
        return "table instance <table: {}, resource: {}>".format(
            self._table_id, self._resource
        )

    def __str__(self) -> str:
        return "table instance <table: {}, resource: {}>".format(
            self._table_id, self._resource
        )

    def _meta(self, table_id: str) -> TableMeta:
        # TODO: catch http bad status
        data = self._requster.table_meta()

        field_metas = {
            field_id: FieldMeta(field_info)
            for field_id, field_info in data["fields"].items()
        }

        return TableMeta(
            table_id=table_id,
            requster=self._requster,
            resource=self._resource,
            field_metas=field_metas,
            field_name_type=self._field_name_type,
        )

    def meta(self) -> TableMeta:
        """get table meta information, like fields

        Returns:
            TableMeta: table meta information
        """
        return self._meta(self._table_id)

    def create(
        self, record: Dict[str, Any], on_conflict: List[str] | str | None = None
    ) -> Record:
        """create single record

        Args:
            record (Dict[str, Any]): record information, key is field name, value is field value
            on_conflict (List[str] | str | None, optional): not create if the field value exist. Default None.

        Returns:
            Record: _description_
        """
        record_values = dict()
        for key, value in record.items():
            record_values[key] = value
        params: Dict[str, Any] = {
            "record": record_values,
        }
        if on_conflict:
            if isinstance(on_conflict, str):
                params["on_conflict"] = [on_conflict]
            elif isinstance(on_conflict, list):
                params["on_conflict"] = on_conflict
            params["action"] = "create_if_not_exists"

        data = self._requster.create_record(
            params,
        )
        new_record_data = data["record"]

        # TODO: add init value
        return Record(
            requester=self._requster,
            record_id=new_record_data["record_id"],
            fields=new_record_data["fields"],
            create_time=new_record_data.get("create_time", None),
            update_time=new_record_data.get("update_time", None),
        )

    def upsert(
        self, record: Dict[str, Any], on_conflict: List[str] | str | None = None
    ) -> Record:
        """upsert single record

        Args:
            record (Dict[str, Any]): record information
            on_conflict (List[str] | str | None, optional): not create if the field value exist. Default None.

        Raises:
            KeyError: field not found

        Returns:
            Record: Record instance
        """

        record_values = dict()
        for key, value in record.items():
            record_values[key] = value
        params: Dict[str, Any] = {
            "record": record_values,
        }
        if on_conflict:
            if isinstance(on_conflict, str):
                params["on_conflict"] = [on_conflict]
            elif isinstance(on_conflict, list):
                params["on_conflict"] = on_conflict
            params["action"] = "upsert"

        data = self._requster.create_record(
            params,
        )
        new_record_data = data["record"]

        # TODO: add init value
        return Record(
            requester=self._requster,
            record_id=new_record_data["record_id"],
            fields=new_record_data["fields"],
            create_time=new_record_data.get("create_time", None),
            update_time=new_record_data.get("update_time", None),
        )

    def bulk_upsert(
        self, records: List[Dict[str, Any]], on_conflict: List[str] | str | None = None
    ) -> Optional[List[Record]]:
        """bulk upsert records

        Args:
            records (List[Dict[str, Any]]): records information
            on_conflict (List[str] | str | None, optional): not create if the field value exist. Default None.

        Returns:
            Optional[List[Record]]: Record instance list
        """
        new_values = []
        for new_record in records:
            record_values = dict()
            for key, value in new_record.items():
                record_values[key] = value
            new_values.append(record_values)

        params: Dict[str, Any] = {
            "records": new_values,
        }
        if on_conflict:
            if isinstance(on_conflict, str):
                params["on_conflict"] = [on_conflict]
            elif is_list_of_strings(on_conflict):
                params["on_conflict"] = on_conflict
        data = self._requster.create_records(
            params,
        )

        if not data or not data["records"]:
            return None

        new_record_datas = data["records"]

        return [
            Record(
                record_id=new_record_data["record_id"],
                fields=new_record_data["fields"],
                requester=self._requster,
                create_time=new_record_data.get("create_time", None),
                update_time=new_record_data.get("update_time", None),
            )
            for new_record_data in new_record_datas
        ]

    def bulk_create(
        self, records: List[Dict[str, Any]], on_conflict: List[str] | str | None = None
    ) -> Optional[List[Record]]:
        """bulk create records

        Args:
            records (List[Dict[str, Any]]): records information
            on_conflict (List[str] | str | None, optional): not create if the field value exist. Default None.

        Raises:
            KeyError: field not found

        Returns:
            Optional[List[Record]]: Record instance list
        """
        new_values = []
        for new_record in records:
            record_values = dict()
            for key, value in new_record.items():
                record_values[key] = value
            new_values.append(record_values)

        params: Dict[str, Any] = {
            "records": new_values,
        }
        if on_conflict:
            if isinstance(on_conflict, str):
                on_conflict = [on_conflict]
            elif isinstance(on_conflict, list):
                params["on_conflict"] = on_conflict
        data = self._requster.create_records(
            params,
        )

        if not data or not data["records"]:
            return None

        new_record_datas = data["records"]

        return [
            Record(
                requester=self._requster,
                record_id=new_record_data["record_id"],
                fields=new_record_data["fields"],
                create_time=new_record_data.get("create_time", None),
                update_time=new_record_data.get("update_time", None),
            )
            for new_record_data in new_record_datas
        ]

    def _table_view2records(self, table_view: Dict[str, Any]) -> List[Record]:
        return [
            Record(
                requester=self._requster,
                record_id=record,
            )
            for record in table_view["records"]
        ]

    def get_by_id(
        self,
        id: str,
    ) -> Optional[Record]:
        """get record by id

        Args:
            id (str): record id

        Returns:
            Optional[Record]: Record instance
        """
        data = self._requster.get_record(
            record_id=id,
        )
        if not data or not data["record"]:
            return None

        return Record(
            requester=self._requster,
            record_id=data["record"]["record_id"],
            fields=data["record"]["fields"],
            create_time=data["record"].get("create_time", None),
            update_time=data["record"].get("update_time", None),
        )

    def get(
        self,
        conditions: Dict[str, Any],
        orders: List[Tuple[str, str]] = [],
    ) -> Optional[Record]:
        """get record by conditions

        Args:
            conditions (Dict[str, Any]): conditions
            orders (List[Tuple[str, str]], optional): order by. Defaults to [].

        Returns:
            Optional[Record]: Record instance list
        """
        where_conditions = [
            LeapcellField(field) == val for field, val in conditions.items()
        ]
        result = KaithQuery(
            self._requster,
            fields=[],
            filter=reduce(lambda x, y: x and y, where_conditions),
            orders=orders,
            limit=1,
        ).query()
        return result[0] if len(result) > 0 else None

    def select(
        self,
        fields: List[str] = [],
    ) -> KaithQuery:
        """get query instance

        Args:
            fields (List[str], optional): fields required. Defaults to [].

        Returns:
            KaithQuery: query instance
        """
        return KaithQuery(self._requster, fields=fields)

    def delete(
        self,
        conditions: Dict[str, Any],
    ) -> int:
        """delete records by conditions

        Args:
            conditions (Dict[str, Any]): conditions

        Returns:
            int: delete count
        """
        filter = KaithQuery(self._requster)._condition2filter(conditions)
        return KaithQuery(self._requster, filter=filter).delete()

    def delete_by_id(
        self,
        id: str,
    ) -> int:
        """delete record by id

        Args:
            id (str): record id

        Returns:
            int: delete count
        """
        data = self._requster.delete_record(
            record_id=id,
        )
        return True

    def count(
        self,
        conditions: Optional[Dict[str, Any] | LeapcellFilter] = None,
    ) -> int:
        """count records by conditions

        Args:
            conditions (Dict[str, Any] | LeapcellFilter, optional): conditions. Defaults to None.

        Returns:
            int: count
        """
        filter = KaithQuery(self._requster)._condition2filter(conditions)
        return KaithQuery(self._requster, filter=filter).count()

    def search(
        self,
        query: str,
        search_fields: List[str] = [],
        fields: List[str] = [],
        boost_fields: Dict[str, int] = {},
        offset: int = 0,
        limit: int = 10,
        conditions: Optional[Dict[str, Any] | LeapcellFilter] = None,
        orders: List[Tuple[str, str]] = [],
    ):
        """_summary_

        Args:
            query (str): the keyword to search
            search_fields (List[str], optional): the fields to search. Defaults to [].
            fields (List[str], optional): the fields to return. Defaults to [].
            boost_fields (Dict[str, int], optional): the fields to boost. Defaults to {}.
            offset (int, optional): offset. Defaults to 0.
            limit (int, optional): limit. Defaults to 10.
            conditions (Optional[Dict[str, Any]  |  LeapcellFilter], optional): conditions. Defaults to None.
            orders (List[Tuple[str, str]], optional): order by. Defaults to [].

        Returns:
            _type_: record instance list
        """
        data = KaithQuery(
            self._requster,
            fields=fields,
            offset=offset,
            limit=limit,
            filter=KaithQuery(self._requster)._condition2filter(conditions),
            orders=orders,
        ).search(query, search_fields, boost_fields)
        return data

    def upload_file(
        self,
        file: bytes,
        filename: Optional[str] = None,
    ) -> LeapcellFile:
        """upload file

        Args:
            file (bytes): file bytes(base64)

        Returns:
            LeapcellFile: file instance
        """
        data = self._requster.upload(
            data=file,
            filename=filename,
        )
        assert isinstance(data, LeapcellFile)
        return data

    def upload_files(
        self,
        files: List[bytes],
    ) -> List[LeapcellFile]:
        """upload files

        Args:
            files (List[bytes]): files bytes(base64), list length must be less than 10

        Returns:
            List[LeapcellFile]: file instance list
        """
        data = self._requster.upload(
            data=files,
        )
        assert isinstance(data, list)
        return data

    def __getitem__(
        self,
        key: str,
    ) -> LeapcellField:
        return LeapcellField(key)

    def toJSON(self):
        meta_info = self.meta()
        return meta_info.toJSON()

    def tojson(self):
        return self.toJSON()



from typing import List, Tuple, Dict, Union, Any
from io import BytesIO
from datetime import datetime
from abc import abstractmethod


class BaseItem:
    def __init__(self) -> None:
        self._data: Any = None
        return

    def to_obj(self) -> Any:
        return self._data

    @property
    def data(self) -> Any:
        return self._data

    @abstractmethod
    def from_obj(self, data: Any) -> None:
        return

    @abstractmethod
    def from_val(self, data: Any) -> None:
        return

    def _process_null_data(self, data: Dict) -> bool:
        if "null" in data and data["null"]:
            self._data = data
            return True
        return False


class StrItem(BaseItem):
    def __init__(self) -> None:
        super().__init__()

    def from_obj(self, data: Any) -> None:
        if isinstance(data, str):
            self._data = data
        else:
            raise ValueError(
                "invalid type {}, str should be str".format(type(data))
            )

    def from_val(self, data: str) -> None:
        if not isinstance(data, str):
            raise TypeError("invalid type {}, it should be str".format(type(data)))
        self._data = data


class FloatItem(BaseItem):
    def __init__(self) -> None:
        super().__init__()

    def from_obj(self, data: Any) -> None:
        if isinstance(data, float):
            self._data = data
        elif isinstance(data, int):
            self._data = float(data)
        else:
            raise ValueError(
                "invalid type {}, float should be float".format(type(data))
            )

    def from_val(self, data: float | int) -> None:
        if not isinstance(data, float) and not isinstance(data, int):
            raise TypeError("invalid type {}, it should be float".format(type(data)))
        self._data = float(data)

class IntItem(BaseItem):
    def __init__(self) -> None:
        super().__init__()

    def from_obj(self, data: Any) -> None:
        if isinstance(data, int):
            self._data = data
        elif isinstance(data, float):
            self._data = int(data)
        else:
            raise ValueError(
                "invalid type {}, int should be int".format(type(data))
            )

    def from_val(self, data: int) -> None:
        if not isinstance(data, int):
            raise TypeError("invalid type {}, it should be int".format(type(data)))
        self._data = data


class LinkItem(BaseItem):
    def __init__(self) -> None:
        super().__init__()

    def from_obj(self, data: Any) -> None:
        if isinstance(data, str):
            self._data = data
        else:
            raise ValueError(
                "invalid type {}, link should be str".format(type(data))
            )

    def from_val(self, data: str) -> None:
        if not isinstance(data, str):
            raise TypeError("invalid type {}, it should be str".format(type(data)))
        self._data = data


class LabelItem(BaseItem):
    def __init__(self) -> None:
        super().__init__()

    def from_obj(self, data: Any) -> None:
        if isinstance(data, str):
            self._data = data
        else:
            raise ValueError(
                "invalid type {}, label should be str".format(type(data))
            )

    def from_val(self, data: str) -> None:
        if not isinstance(data, str):
            raise TypeError("invalid type {}, it should be str".format(type(data)))
        self._data = data


class LabelsItem(BaseItem):
    def __init__(self) -> None:
        super().__init__()

    def from_obj(self, data: Any) -> None:
        if isinstance(data, list):
            self._data = data
        else:
            raise ValueError(
                "invalid type {}, labels should be list of str".format(type(data))
            )

    def from_val(self, data: List[str]) -> None:
        if not isinstance(data, list):
            raise TypeError("invalid type {}, it should be list".format(type(data)))
        self._data = data


class ImageItem(BaseItem):
    def __init__(self) -> None:
        super().__init__()

    def from_obj(self, data: Any) -> None:
        if isinstance(data, str):
            self._data = data
        else:
            raise ValueError(
                "invalid type {}, image should be str".format(type(data))
            )

    def from_val(self, data: str | BytesIO) -> None:
        if not isinstance(data, str):
            raise TypeError("invalid type {}, it should be str".format(type(data)))
        self._data = data
        return


class ImagesItem(BaseItem):
    def __init__(self) -> None:
        super().__init__()

    def from_obj(self, data: Any) -> None:
        if isinstance(data, list):
            self._data = data
        else:
            raise ValueError(
                "invalid type {}, images should be list".format(type(data))
            )

    def from_val(self, data: List[str]) -> None:
        if not isinstance(data, list):
            raise TypeError("invalid type {}, it should be list".format(type(data)))
        self._data = data


class TimeItem(BaseItem):
    def __init__(self) -> None:
        super().__init__()

    def from_obj(self, data: Any) -> None:
        if isinstance(data, int): #TODO: datetime?
            self._data = data
            self._raw = data
        else:
            raise ValueError(
                "invalid type {}, time should be int".format(type(data))
            )

    def from_val(self, data: int | datetime) -> None:
        if not isinstance(data, int) and not isinstance(data, datetime):
            raise TypeError("invalid type {}, it should be int".format(type(data)))
        if isinstance(data, int):
            self._data = data
        elif isinstance(data, datetime):
            self._data = datetime.timestamp(data)
        return


class LongTextItem(StrItem):
    def __init__(self) -> None:
        super().__init__()


class FieldMgr:
    def __init__(self) -> None:
        pass

    @staticmethod
    def from_obj(fieldType: str, data: Dict) -> BaseItem:
        if data is None:
            return None
        item = FieldMgr._get_item(fieldType=fieldType)
        item.from_obj(data=data)
        return item

    @staticmethod
    def from_val(fieldType: str, data: Any) -> BaseItem:
        item = FieldMgr._get_item(fieldType=fieldType)
        item.from_val(data=data)
        return item

    @staticmethod
    def _get_item(fieldType: str):
        item: Union[
            None,
            IntItem,
            StrItem,
            FloatItem,
            TimeItem,
            LinkItem,
            LabelItem,
            LabelsItem,
            ImagesItem,
            ImageItem,
            LongTextItem,
        ] = None
        if fieldType == "INT_NUMBER":
            item = IntItem()
        elif fieldType == "STR":
            item = StrItem()
        elif fieldType == "FLOAT_NUMBER":
            item = FloatItem()
        elif fieldType == "TIME":
            item = TimeItem()
        elif fieldType == "LINK":
            item = LinkItem()
        elif fieldType == "LABEL":
            item = LabelItem()
        elif fieldType == "LABELS":
            item = LabelsItem()
        elif fieldType == "IMAGES":
            item = ImagesItem()
        elif fieldType == "IMAGE":
            item = ImageItem()
        elif fieldType == "LONG_TEXT":
            item = LongTextItem()
        else:
            raise KeyError("not found type '{}'".format(fieldType))
        return item



import httpx
from typing import Dict, Any, Union, List, Optional, Tuple
import os
from leapcell.exp import LeapcellException
from leapcell.version import VERSION
import urllib.parse
from functools import reduce
from leapcell.utils import multi_urljoin, build_header
from leapcell.file import LeapcellFile
from io import BytesIO
import urllib3
import json

try:
    from urllib.parse import urlparse, ParseResult
except ImportError:
    from urlparse import urlparse, ParseResult  # type: ignore

MAX_CONNECTION_RETRIES = 2
TIMEOUT_SECS = 600
FILE_UPLOAD_TIMEOUT = 600
FILE_UPLOAD_MAX_SIZE = 1024 * 1024 * 3


def endpoint(resource: str, table_id: str, version="v1", name_type="id") -> str:
    return multi_urljoin("/api/", version + "/", resource + "/", "/table/", table_id)


class HTTPClient(object):
    def __init__(
        self,
        api_key: str,
        base_url: str,
        resource: str,
        table_id: str,
        version="v1",
        name_type="id",
    ) -> None:
        self._base_url = base_url
        self._api_key = api_key
        self._url_prefix = endpoint(
            resource, table_id, version=version, name_type=name_type
        )
        self._table_id = table_id
        self._name_type = name_type

    def _request(
        self,
        url_path: str,
        method: str,
        data: Optional[Union[Dict[str, Any], List[Dict[str, Any]]]] = None,
        params: Optional[Dict[str, Any]] = None,
        files: Optional[
            Dict[str, Tuple[str, BytesIO]] | List[Tuple[str, BytesIO]]
        ] = None,
    ) -> Any:
        # use urllib3 instead of requests to avoid requests dependency
        client = urllib3.PoolManager()
        urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

        url = urllib.parse.urljoin(self._base_url, url_path)
        if params is not None:
            query_string = urllib.parse.urlencode(params)
            url = url + "?" + query_string

        fields = None
        if isinstance(files, dict):
            fields = {
                "file": (files["file"][0], files["file"][1].getbuffer().tobytes()),
            }
        elif isinstance(files, list):
            fields = [("files", (f[0], f[1].getbuffer().tobytes())) for f in files]

        body = None
        if data is not None:
            body = json.dumps(data)

        headers = urllib3.HTTPHeaderDict()
        for k, v in build_header(self._api_key).items():
            headers.add(k, v)
        headers.add(
            "Accept-Encoding", "gzip"
        )  # use gzip, urllib3 will decompress automatically

        response = client.request(
            method=method,
            url=url,
            headers=headers,
            timeout=TIMEOUT_SECS,
            retries=urllib3.Retry(MAX_CONNECTION_RETRIES, redirect=2),
            fields=fields,
            body=body,
        )
        print(response.headers)
        try:
            body_json = json.loads(response.data)
        except ValueError:
            raise LeapcellException(
                "bad response, body is not json, http code {}, body: {}".format(
                    response.status, response.data
                )
            )
        if response.status != 200:
            raise LeapcellException(
                "bad request, http code {}, please check apitoken and params, error code: {}, hint: {}".format(
                    response.status,
                    body_json.get("code", ""),
                    body_json.get("error", ""),
                )
            )

        if "data" not in body_json:
            raise LeapcellException(
                "bad request, http code {}, please check apitoken and params, error code: {}, hint: {}".format(
                    response.status,
                    body_json.get("code", ""),
                    body_json.get("error", ""),
                )
            )
        return body_json["data"]

    def table_meta(self) -> Any:
        name_type = self._name_type
        return self._request(
            url_path="{}".format(self._url_prefix),
            method="GET",
            params={
                "name_type": name_type,
            },
        )

    def create_record(self, data: Dict) -> Dict[str, Any]:
        data["name_type"] = self._name_type
        return self._request(
            url_path="{}/record".format(self._url_prefix),
            method="POST",
            data=data,
        )

    def create_records(self, data: Dict[str, Any]) -> Optional[Dict[str, Any]]:
        data["name_type"] = self._name_type
        return self._request(
            url_path="{}/record".format(self._url_prefix),
            method="POST",
            data=data,
        )

    def get_record(self, record_id: str) -> Optional[Dict[str, Any]]:
        return self._request(
            url_path="{}/record/{}".format(self._url_prefix, record_id),
            method="GET",
            params={
                "name_type": self._name_type,
            },
        )

    def get_records(self, data: Dict[str, Any]) -> Optional[Dict[str, Any]]:
        data["name_type"] = self._name_type
        return self._request(
            url_path="{}/record/query".format(self._url_prefix),
            method="POST",
            data=data,
        )

    def update_records(self, data: Dict[str, Any]) -> Optional[Dict[str, Any]]:
        data["name_type"] = self._name_type
        return self._request(
            url_path="{}/record".format(self._url_prefix),
            method="PUT",
            data=data,
        )

    def update_record(
        self, record_id: str, data: Dict[str, Any]
    ) -> Optional[Dict[str, Any]]:
        data["name_type"] = self._name_type
        return self._request(
            url_path="{}/record/{}".format(self._url_prefix, record_id),
            method="PUT",
            data=data,
        )

    def delete_records(self, data: Dict[str, Any]) -> Optional[Dict[str, Any]]:
        data["name_type"] = self._name_type
        return self._request(
            url_path="{}/record".format(self._url_prefix),
            method="DELETE",
            data=data,
        )

    def delete_record(self, record_id) -> Optional[Dict[str, Any]]:
        return self._request(
            url_path="{}/record/{}".format(self._url_prefix, record_id),
            method="DELETE",
        )

    def aggr_record(self, data: Dict[str, Any]) -> Optional[Dict[str, Any]]:
        data["name_type"] = self._name_type
        return self._request(
            url_path="{}/record/metrics".format(self._url_prefix),
            method="POST",
            data=data,
        )

    def search(self, data: Dict[str, Any]) -> Optional[Dict[str, Any]]:
        data["name_type"] = self._name_type
        return self._request(
            url_path="{}/record/search".format(self._url_prefix),
            method="POST",
            data=data,
        )

    def upload(
        self, data: bytes | List[bytes], filename: Optional[str] = None
    ) -> LeapcellFile | List[LeapcellFile]:
        if isinstance(data, bytes):
            return self._upload(BytesIO(data), filename)
        elif isinstance(data, list):
            return self._upload_multi([BytesIO(d) for d in data])
        else:
            raise TypeError("invalid data {}, which should be byteIO".format(data))

    def _upload(self, file: BytesIO, filename: Optional[str]) -> LeapcellFile:
        if file.getbuffer().nbytes > FILE_UPLOAD_MAX_SIZE:
            raise LeapcellException("file is too large, file should be less than 5KB")
        if filename is not None:
            files = {"file": (filename, file)}
        else:
            files = {"file": ("file", file)}
        r = self._request(
            url_path="{}/{}".format(self._url_prefix, "upload"),
            method="POST",
            files=files,
        )
        response = r
        image_item = LeapcellFile(response.get("file", {}))
        return image_item

    def _upload_multi(self, files: List[BytesIO]) -> List[LeapcellFile]:
        upload_files = []
        for f in files:
            if f.getbuffer().nbytes > FILE_UPLOAD_MAX_SIZE:
                raise LeapcellException(
                    "file is too large, file should be less than 5KB"
                )
            upload_files.append(("files", (f)))
        r = self._request(
            url_path="{}/{}".format(self._url_prefix, "upload_multi"),
            method="POST",
            files=upload_files,
        )
        response = r
        image_item = [LeapcellFile(item) for item in response.get("files", [])]
        return image_item



from typing import Dict


class FieldMeta:
    def __init__(self, obj: Dict) -> None:
        self._id: str = obj["id"]
        self._type: str = obj["type"]
        self._name: str = obj["name"]
        pass

    @property
    def type(self):
        return self._type

    @property
    def id(self):
        return self._id

    @property
    def name(self):
        return self._name

    def __str__(self) -> str:
        return "<field id: {} name: {}, type: {}>".format(
            self.id,
            self.name,
            self.type,
        )

    def __repr__(self) -> str:
        return self.__str__()
    
    def todict(self) -> Dict[str, str]:
        return {
            "id": self.id,
            "name": self.name,
            "type": self.type,
        }
    
    def toJSON(self):
        return self.todict()



from setuptools import setup, find_packages

setup(
    name="leapcell",
    keywords=["back-end", "framework", "RESTful"],
    version="0.0.5",
    description="leapcell python client",
    author="aljun",
    author_email="gagasalamer@outlook.com",
    license="Apache",
    url="https://docs.leapcell.io",
    install_requires=[
        "httpx >= 0.15.0",
    ],
    python_requires=">=3.5",
    packages=["leapcell"],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



from leapcell.table import LeapcellTable, LeapcellField, LeapcellFilter
import os
from typing import List, Tuple, Dict, Union


class Leapcell(object):
    """Leapcell is the main class of Leapcell SDK, it is used to create table client and config client."""

    default_base_url = "https://api.leapcell.io"

    def __init__(
        self, api_key: str, base_url: str | None = None, version: str = "v1"
    ) -> None:
        """_summary_

        Args:
            api_key (str): Bearer token for authentication, api_key is the api token of your Leapcell account, you can find it in your account page: leapcell.io/account.
            base_url (str, optional): base_url is the url of your Leapcell server, if you use the Leapcell cloud service, you can ignore this parameter. Defaults to None.
            version (str, optional): version is the version of Leapcell server, if you use the Leapcell cloud service, you can ignore this parameter. Defaults to "v1".

        Raises:
            Exception: api_key can not be empty, you can find it in your account page: leapcell.io/account
        """
        self._base_url = (
            base_url or os.environ.get("LEAPCELL_API_URL") or self.default_base_url
        )
        self._api_key = api_key or os.environ.get("LEAPCELL_API_TOKEN")
        if not api_key:
            raise Exception(
                "api_key can not be empty, you can find it in your account page: leapcell.io/account"
            )
        self._version = version

    def table(
        self, repository: str, table_id: str, name_type: str = "name"
    ) -> LeapcellTable:
        """table is used to create table instance.

        Args:
            repository (str): leapcell repository, you can find it in leapcell with format leapcell.io/{username}/{repo_name}, resource all have the same format as {username}/{repo_name}
            table_id (str): table id, you can find it in leapcell with format leapcell.io/{username}/{repo_name}/table/{table_id}, table_id all have the same format as {username}/{repo_name}/table/{table_id}
            name_type (str, optional): name_type is the type of the field, it can be "id" or "name". Defaults to "name". if it's "id", the field name will be "13145252145", "13145252147" ...; if it's "name", the field name will be "field-0", "field-1", "field-2", ...,

        Raises:
            ValueError: resource can not be empty
            ValueError: table_id can not be empty
            ValueError: api_key can not be empty

        Returns:
            LeapcellTable: LeapcellTable instance
        """
        if not repository:
            raise ValueError("repository can not be empty")
        if not table_id:
            raise ValueError("table_id can not be empty")
        if not self._api_key:
            raise ValueError("api_key can not be empty")

        return LeapcellTable(
            repository=repository,
            api_key=self._api_key,
            table_id=table_id,
            base_url=self._base_url,
            name_type=name_type,
            version=self._version,
        )



from leapcell import Leapcell
import requests

# init client, with token xxx
leapclient = Leapcell("lpcl_1566404210.efc9ad39ddf7bc4eea34275a350fd354")


# init table instance
table = leapclient.table("salamer/myblog", "tbl1700559237082210304", name_type="name")

# get table fields
print(table.meta().toJSON())

# create records
record = table.create(
    {
        "name": "cyy",
    }
)

count = table.count(
    {
        "name": "cyy",
    }
)
print(count)

# update record
record["name"] = "hello"
record.save()

# # get record by id
record = table.get_by_id(record.id)
print(record.toJSON())

count = table.count(
    {
        "name": "cyy",
    }
)
print(count)

# # get records
records = table.select().where({
    "name": "cyy",
}).limit(10).offset(0).first()
print(records)

# # update records
table.select().where({
    "name": "cyy",
}).update({
    "name": "hello",
})

# delete record
record.delete()





image_req = requests.get("https://leapcell-dev-bucket.s3.amazonaws.com/920417ee6a70435084fe6b40d58dad4e.jpeg")
image = image_req.content
resp = table.upload_file(image)
# # LeapcellImage(id=img46fa2754a8fa4719a3cb66db06b25e42-upload, link=https://leapcell-dev-bucket.s3.amazonaws.com/img46fa2754a8fa4719a3cb66db06b25e42-upload, width=600, height=300)
print(resp)
resp = table.upload_files([image, image])
print(resp)


from typing import Dict
from leapcell.field_meta import FieldMeta
from leapcell.http_client import HTTPClient
from leapcell.const import TableFieldType
import json


class TableMeta():
    """Leapcell Table Meta Class
    """
    def __init__(self,
                 resource: str,
                 table_id: str,
                 requster: HTTPClient,
                 field_metas: Dict[str, FieldMeta],
                 field_name_type: TableFieldType) -> None:
        self._requster = requster
        self._resource = resource
        self._table_id = table_id
        self._field_metas = field_metas
        self._display_field_metas = {
            item.name: item for item in field_metas.values()}
        self._field_name_type = field_name_type

    @property
    def requster(self):
        return self._requster

    @property
    def resource(self):
        return self._resource

    @property
    def table_id(self):
        return self._table_id

    @property
    def field_metas(self):
        if self._field_name_type == TableFieldType.ID:
            return self._field_metas
        elif self._field_name_type == TableFieldType.NAME:
            return self._display_field_metas
        return self._field_metas
    
    @property
    def field_id_metas(self):
        return self._field_metas

    @property
    def display_field_metas(self):
        return self._display_field_metas

    @property
    def field_name_type(self):
        return self._field_name_type

    def __str__(self) -> str:
        return "<repo: {}, table: {}>".format(
            self._resource, self.table_id)

    def __repr__(self) -> str:
        return json.dumps(self.toJSON())
    
    def todict(self) -> Dict[str, FieldMeta]:
        return {k:v for k,v in self.field_metas.items()}
    

    def toJSON(self):
        return {
            "resource": self._resource,
            "table_id": self._table_id,
            "fields": self.todict(),
        }
    
    def tojson(self):
        return self.toJSON()